



python -c "import tiledb; tiledb.libtiledb.version()"
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -c "import tiledb; assert tiledb.VFS().supports('s3')"
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest -s -v --ignore-glob='*_examples.py' --ignore-glob='*_dask.py'
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
